# Copyright (C) 2010 W. Trevor King <wking@drexel.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""An email interface for players.
"""

from __future__ import absolute_import
from cStringIO import StringIO
from email.generator import Generator
from email.parser import Parser
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import getaddresses, formataddr, formatdate, make_msgid
import os
import smtplib
import subprocess
import sys
import tempfile

from ..base import Player, PlayerError


# Configure alternative sendmail command in case smtplib is too
# annoying.  Set SENDMAIL to None to use smtplib.
SENDMAIL = None
#SENDMAIL = ['/usr/sbin/sendmail', '-t']
#SENDMAIL = ['/usr/bin/msmtp', '-t']


class IncomingEmailDispatcher (object):
    """For reading reply messages.
    """
    def __init__(self, fifo_path=None, verbose=True):
        self.verbose = verbose
        self._cache = []
        if fifo_path == None:
            self.dir_path = tempfile.mkdtemp(suffix='.pyrisk')
            self.fifo_path = os.path.join(self.dir_path, 'incoming')
        else:
            self.dir_path = None
            self.fifo_path = os.path.abspath(fifo_path)
        os.mkfifo(self.fifo_path)
    def close(self):
        os.remove(self.fifo_path)
        if self.dir_path != None:
            os.rmdir(self.dir_path)
    def get(self, tag):
        for msg_tag, msg in self._cache:
            if msg_tag == tag:
                self._cache.remove(msg)
                return msg
        msg = self._get_msg()
        msg_tag = self._msg_tag(msg['Subject'])
        while msg_tag != tag:
            self._cache.append((msg_tag, msg))
            msg = self._get_msg()
            msg_tag = self._msg_tag(msg['Subject'])
        if self.verbose == True:
            print >> sys.stderr, msg
        return msg
    def _msg_tag(self, subject):
        """ Return the tag portion of a message subject.

        >>> ied = IncomingEmailDispatcher()
        >>> ied._msg_tag('[TAG] Hi there')
        u'[TAG]'
        >>> ied._msg_tag('  [tg] Hi there')
        u'[tg]'
        >>> ied._msg_tag(' Re: [t] Hi there')
        u'[t]'
        >>> ied.close()
        """
        if subject == None:
            return None
        subject = subject.strip()
        if subject.startswith('Re:'):
            subject = subject[len('Re:'):]
            subject = subject.strip()
        args = subject.split(u']',1)
        if len(args) < 1:
            return None
        return args[0]+u']'
    def _get_msg(self):
        # FIFO blocks on open until a writer also opens
        self.fifo = open(self.fifo_path, 'r')
        text = self.fifo.read()
        self.fifo.close()
        p = Parser()
        return p.parsestr(text)

class OutgoingEmailDispatcher (object):
    """For sending outgoing messages.
    """
    def __init__(self, return_address, return_name='PyRisk server',
                 sendmail=None, verbose_execute=False,
                 smtp_host=None, smtp_port=465,
                 smtp_user=None, smtp_password=None):
        self.return_address = return_address
        self.return_name = return_name
        self.sendmail = sendmail
        if self.sendmail == None:
            self.sendmail = SENDMAIL
        self.verbose_execute = verbose_execute
        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_user = smtp_user
        self.smtp_password = smtp_password
    def send(self, msg):
        """Send an email Message instance on its merry way.
        """
        msg['From'] = formataddr((self.return_name, self.return_address))
        msg['Reply-to'] = msg['From']
        msg['Date'] = formatdate()
        msg['Message-id'] = make_msgid()

        if self.sendmail != None:
            self._execute(self.sendmail, stdin=self._flatten(msg))
            return None
        s = smtplib.SMTP_SSL(self.smtp_host, self.smtp_port)
        s.connect()
        s.login(self.smtp_user, self.smtp_password)
        s.sendmail(from_addr=self._source_email(msg),
                   to_addrs=self._target_emails(msg),
                   msg=self._flatten(msg))
        s.close()
    def _execute(self, args, stdin=None, expect=(0,)):
        """
        Execute a command (allows us to drive gpg).
        """
        if self.verbose_execute == True:
            print >> sys.stderr, '$ '+args
        try:
            p = subprocess.Popen(args,
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 shell=False, close_fds=True)
        except OSError, e:
            strerror = '%s\nwhile executing %s' % (e.args[1], args)
            raise Exception, strerror
        output, error = p.communicate(input=stdin)
        status = p.wait()
        if self.verbose_execute == True:
            print >> sys.stderr, '(status: %d)\n%s%s' % (status, output, error)
        if status not in expect:
            strerror = '%s\nwhile executing %s\n%s\n%d' % (args[1], args, error, status)
            raise Exception, strerror
        return status, output, error
    def _source_email(self, msg, return_realname=False):
        """
        Search the header of an email Message instance to find the
        sender's email address.
        """
        froms = msg.get_all('from', [])
        from_tuples = getaddresses(froms) # [(realname, email_address), ...]
        if return_realname == True:
            return from_tuples[0] # (realname, email_address)
        return from_tuples[0][1]  # email_address    
    def _target_emails(self, msg):
        """
        Search the header of an email Message instance to find a
        list of recipient's email addresses.
        """
        tos = msg.get_all('to', [])
        ccs = msg.get_all('cc', [])
        bccs = msg.get_all('bcc', [])
        resent_tos = msg.get_all('resent-to', [])
        resent_ccs = msg.get_all('resent-cc', [])
        resent_bccs = msg.get_all('resent-bcc', [])
        resent = resent_tos + resent_ccs + resent_bccs
        if len(resent) > 0:
            all_recipients = getaddresses(resent)
        else:
            all_recipients = getaddresses(tos + ccs + bccs)
        return [addr[1] for addr in all_recipients]
    def _flatten(self, msg, to_unicode=False):
        """
        Produce flat text output from an email Message instance.
        """
        assert msg != None
        fp = StringIO()
        g = Generator(fp, mangle_from_=False)
        g.flatten(msg)
        text = fp.getvalue()
        if to_unicode == True:
            encoding = msg.get_content_charset('utf-8')
            text = unicode(text, encoding=encoding)
        return text

def encodedMIMEText(body, encoding='us-ascii', disposition='inline', filename=None):
    if encoding == 'us-ascii':
        part = MIMEText(body)
    else:
        # Create the message ('plain' stands for Content-Type: text/plain)
        part = MIMEText(body.encode(encoding), 'plain', encoding)
    if filename == None:
        part.add_header('Content-Disposition', disposition)
    else:
        part.add_header('Content-Disposition', disposition, filename=filename)
    return part


class EmailPlayer (Player):
    """Human Player with an email interface.

    TODO: details on procmail setup
    """
    def __init__(self, name, address, incoming, outgoing, world_renderer=None):
        Player.__init__(self, name)
        self.address = address
        self.outgoing = outgoing
        self.incoming = incoming
        self.world_renderer = world_renderer
    def _tag(self):
        return '[PyRisk %d]' % (id(self))
    def _send_mail(self, world, log, subject, body):
        msg = MIMEMultipart()
        msg.attach(encodedMIMEText(body, filename='body'))
        msg.attach(self._log_part(log))
        msg.attach(self._world_part(world))
        if self.world_renderer != None:
            msg.attach(self._rendered_world_part(world, log))
        msg['To'] = formataddr((self.name, self.address))
        tag = self._tag()
        msg['Subject'] = '%s %s' % (tag, subject)
        self.outgoing.send(msg)
        return tag
    def _get_mail(self, tag):
        msg = self.incoming.get(tag)
        msg_charset = msg.get_content_charset('utf-8')
        if msg.is_multipart():
            for part in msg.walk():
                mime_type = part.get_content_type()
                if mime_type == 'text/plain':
                    break
            body = part.get_payload(decode=True)
            charset = part.get_content_charset(msg_charset)
        else:
            body = msg.get_payload(decode=True)
            charset = msg_charset
            mime_type = msg.get_content_type()
        if not mime_type.startswith('text/plain'):
            raise PlayerError('Invalid MIME type %s (must be text/plain)'
                              % mime_type)
        body = unicode(body, charset) # convert text types to unicode
        if len(body) == 0 or body[-1] != u'\n':
            body += u'\n'
        return body
    def _world_part(self, world):
        body = []
        for continent in world:
            body.append(str(continent))
            for terr in continent:
                if terr.player == None:
                    body.append('  %s\t%s' % (terr, terr.player))
                else:
                    body.append('  %s\t%s\t%d' % (terr, terr.player, terr.armies))
        return encodedMIMEText('\n'.join(body), filename='world')
    def _rendered_world_part(self, world, log):
        start_event = log[0]
        players = start_event.players
        body = self.world_renderer.render(world, players)
        filename,subtype = \
            self.world_renderer.filename_and_mime_image_type(world)
        part = MIMEImage(body, subtype)
        part.add_header('Content-Disposition', 'attachment', filename=filename)
        return part
    def _log_part(self, log):
        return encodedMIMEText('\n'.join([str(e) for e in log]),
                               filename='log')
    def report(self, world, log):
        """Send reports about death and game endings.

        These events mark the end of contact and require no change in
        player status or response, so they get a special command
        seperate from the usual action family.  The action commands in
        Player subclasses can notify the player (possibly by calling
        report internally) if they feel so inclined.
        
        See also
        --------
        draw - another notification-only method
        """
        self._send_mail(world, log, 'Report: %s' % log[-1],
                        Player.report(world, log))
    def draw(self, world, log, cards=[]):
        """Only called if you earned a new card (or cards).

        See also
        --------
        report - another notification-only method
        """
        Player.draw(self, world, log, cards)
        body = ['New cards:']
        body.extend(['  %s' % c for c in cards])
        body = ['Current Hand:']
        for c in self.hand:
            if c.territory != None and c.territory.player == self:
                body.append('  %s (owned)' % c)
            else:
                body.append('  %s' % c)
        self._send_mail(world, log, 'Drawing cards', '\n'.join(body))
    def __select_territory(self, world, log, error=None):
        """Return the selected territory's name.
        """
        body = [
            'Reply with first line of the body of your email set',
            'to the name (long or short, case insenitive) of the',
            'territory you wish to occupy.  Available territories',
            'are:']
        for t in world.territories():
            if t.player == None:
                body.append('  %s' % t)
        if error != None:
            body.insert(0, '')
            body.insert(0, str(error))
        tag = self._send_mail(world, log, 'Select territory', '\n'.join(body))
        body = self._get_mail(tag)
        name = body.splitlines()[0].strip()
        return name
    def play_cards(self, world, log, error=None,
                   play_required=True):
        """Decide whether or not to turn in a set of cards.

        Return a list of cards to turn in or None.  If play_required
        is True, you *must* play.
        """
        possibles = list(self.hand.possible())
        if len(possibles) == 0:
            return None
        subject = 'Play cards'
        if play_required == True:
            subject += ' (required)'
        body = [
            'Reply with first line of the body of your email set',
            'to the number of the set you wish to play (leave body',
            'blank to pass).  Available sets are:']
        for i,h in enumerate(possibles):
            body.append('  %d: %s' % (i, h))
        if error != None:
            body.insert(0, '')
            body.insert(0, str(error))
        tag = self._send_mail(world, log, subject, '\n'.join(body))
        body = self._get_mail(tag)
        text = body.splitlines()[0].strip()
        if text == '':
            return None
        return possibles[int(text)]
    def place_armies(self, world, log, error=None,
                     remaining=1, this_round=1):
        """Both during setup and before each turn.

        Return {territory_name: num_armies, ...}
        """
        subject = 'Place %d of %d armies' % (this_round, remaining)
        body = [
            'You can place %d armies this round (out of %d in'
            % (this_round, remaining),
            'this phase).',
            '',
            'Reply with first line(s) of the body of your email set',
            'to "<number_of_armies> : <territory_name>" followed by',
            'a blank line.  For example',
            '  1 : gbr',
            '  4 : indo',
            '  ...',
            ''
            'Your current disposition is:']
        for t in self.territories(world):
            body.append('  %d : %s' % (t.armies, t))
        if error != None:
            body.insert(0, '')
            body.insert(0, str(error))
        tag = self._send_mail(world, log, subject, '\n'.join(body))
        body = self._get_mail(tag)
        placements = {}
        for line in body.splitlines():
            line = line.strip()
            if len(line) == 0:
                break
            if line.count(':') != 1:
                raise PlayerError('Invalid syntax "%s"' % line)
            armies,terr_name = [x.strip() for x in line.split(':')]
            placements[terr_name] = int(armies)
        return placements
    def attack_and_fortify(self, world, log, error=None,
                           mode='attack'):
        """Return list of (source, target, armies) tuples.  Place None
        in the list to end this phase.
        """
        if mode == 'attack':
            subject = 'Attack and fortify'
            body = [
                'You can attack as many times as you like, and fortify',
                'once at the end of the round.  Reply with first line(s)',
                'of the body of your email set to',
                '  "<source_name> : <target_name> : <number_of_armies>',
                'When you are done attacking or in place of a',
                'fortification, insert the line "Pass".  For example',
                '  gbr : ice : 3',
                '  jap : chi : 3',
                '  jap : chi : 3',
                '  Pass',
                '  gbr : seu : 7',
                '  ',
                'or',
                '  jap : chi : 3',
                '  Pass',
                '  Pass',
                '  ']
        else:
            assert mode == 'fortify', mode
            subject = 'Fortify'
            body = [
                'You can fortify once.  Reply with first line of the',
                'body of your email set to',
                '  "<source_name> : <target_name> : <number_of_armies>',
                'Or, if you choose to pass, either a blank line or',
                '"Pass".  For example',
                '  gbr : seu : 7',
                'or',
                '  ',
                'or',
                '  Pass']
        if error != None:
            body.insert(0, '')
            body.insert(0, str(error))
        tag = self._send_mail(world, log, subject, '\n'.join(body))
        body = self._get_mail(tag)
        if mode == 'fortify':
            return [self._parse_attack_or_fortify_line(
                        body.splitlines()[0], mode)]
        actions = []
        pass_count = 0
        for line in body.splitlines():
            action = self._parse_attack_or_fortify_line(line, mode)
            if action == None:
                pass_count += 1
                if pass_count == 2:
                    break
            actions.append(action)
        return actions
    def _parse_attack_or_fortify_line(self, line, mode):
        line = line.strip()
        if line.count(':') == 2:
            fields = [x.strip() for x in line.split(':')]
            fields[2] = int(fields[2])
            return fields
        elif line.lower() == 'pass' \
                or (mode == 'fortify' and len(line) == 0):
            return None
    def support_attack(self, world, log, error,
                       source, target):
        """Follow up on a conquest by moving additional armies.
        """
        subject = 'Support conquest of %s by %s' % (target, source)
        body = [
            'You can move up to %d of the %d armies remaining on'
            % (source.armies - 1, source.armies),
            '%s to %s following your conquest.'
            % (source, target),
            '',
            'Reply with first line(s) of the body of your email set',
            'to "<number_of_armies>", or leave the first line blank',
            'to pass.']
        if error != None:
            body.insert(0, '')
            body.insert(0, str(error))
        tag = self._send_mail(world, log, subject, '\n'.join(body))
        body = self._get_mail(tag)
        text = body.splitlines()[0].strip()
        if text == '':
            return 0
        return int(text)


def test():
    import doctest
    failures,tests = doctest.testmod(sys.modules[__name__])
    return failures
